"""Module containing the SpurGear class"""
# I want the variables names to be the same as in AGMA pylint: disable=invalid-name
from math import pi, radians, cos, sin

from me_toolbox.gears import Gear


class SpurGear(Gear):
    """A gear object that contains all of its design parameters (AGMA 2001-D04)"""

    # TODO: YN - add solution for low cycle of material not in the graph
    # TODO: Qv - add a way to calculate the Qv value as in AGMA 2001-D04

    def __init__(self, modulus, teeth_num, rpm, Qv, width, bearing_span, pinion_offset, enclosure,
                 hardness, pressure_angle, grade, work_hours=0, number_of_cycles=0, crowned=False,
                 adjusted=False, sensitive_use=False, nitriding=False, case_carb=False,
                 material='steel'):
        """ Instantiating SpurGear object

        :param float modulus: modulus in [mm]
        :param float pressure_angle: pressure angle in [deg]
        :param int teeth_num: number of teeth
        :param float rpm: angular velocity in [RPM]
        :param int grade: material grade (1 or 2)
        :param int Qv: transmission quality [5<=Qv<=11]
        :param bool crowned: crowned teeth shape ( True / False )
        :param bool adjusted: Adjusted after assembly ( True / False )
        :param float width: gear width in [mm]
        :param float bearing_span: length between the middle of the bearings in [mm]
        :param float pinion_offset: gear offset from the middle of the bearing span in [mm]
        :param str enclosure: type of enclosure
                            (open gearing / commercial enclosed / precision enclosed /
                            extra precision enclosed)
        :param float hardness: gear hardness in Brinell scale [HB]
        :param str material: gear material (steel, malleable iron, nodular iron, cast iron,
            aluminum bronze, tin bronze)
        :param float work_hours: number of hour of operation in [hr]
        :param float number_of_cycles: number of cycles of operation
        :param bool sensitive_use: if the gear is for sensitive use ( True / False )
        :param bool nitriding: heat treating process ( True / False )

        :returns: A SpurGear object
        :rtype: SpurGear
        """
        super().__init__(modulus, teeth_num, rpm, Qv, width, bearing_span, pinion_offset, enclosure,
                         hardness, pressure_angle, grade, work_hours, number_of_cycles, crowned,
                         adjusted, sensitive_use, nitriding, case_carb, material)

    def get_info(self):
        """Print all the class fields with values """
        for key in self.__dict__:
            print(f"{key} : {self.__dict__[key]}")

    def get_factors(self, verbose=True):
        """Print correction factors for gear strength analysis

        :param bool verbose: Print factors
        :rtype: dict
        """

        factors = {'factor_KB': self.KB, 'Kv': self.Kv, 'max_vel': self.maximum_velocity,
                   'factor_Ks': self.Ks, 'KH': self.KH, 'St': self.St, 'ZR': self.ZR, 'Sc': self.Sc,
                   'YN': self.YN, 'ZN': self.ZN, 'Yj': self.__dict__.get('Yj', None),
                   'Zw': self.__dict__.get('Zw', None)}

        if verbose:
            print("factor_KB= {KB}, Kv= {Kv}, maximum velocity= {max_vel}, \
                  \nfactor_Ks= {Ks}, KH= {KH}, St= {St}, ZR= {ZR}, Sc= {Sc}, \
                  \nYN= {YN}, ZN= {ZN}, Yj= {Yj}, Zw= {Zw}".format(**factors))

        if None in factors.values():
            print("None value caused by Transmission not been instantiated")

        return factors

    @staticmethod
    def ZI(gear1, gear2):
        """Geometric factor for contact failure,
        ZI is dependent on the gear ratio and pressure angel

        :param SpurGear gear1: gear object
        :param SpurGear gear2: gear object

        :return: ZI - geometric factor
        :rtype: float
        """

        mG = gear2.teeth_num / gear1.teeth_num
        phi = radians(gear1.pressure_angle)
        Z_I = 0.5 * cos(phi) * sin(phi) * (mG / (mG + 1))
        return Z_I

    def optimization(self, transmission, optimize_feature='all', verbose=False):
        """Perform gear optimization

        :param gears.transmission.Transmission transmission: Transmission object
            associated with the gears
        :param str optimize_feature: property to optimize for ('width'/'volume'/'center')
        :param bool verbose: print optimization stages

        :return: optimized result (width in mm, volume in mm^3, center distance in mm)
        :rtype: dict
        """

        gear = self

        # saving original attribute values
        original_teeth_num = gear.teeth_num  # saving original attribute
        original_width = gear.width
        original_modulus = gear.modulus

        # choosing the minimum number of teeth to start with
        # (Note: this is the minimum number of teeth to avoid interference)
        if gear.pressure_angle == 20:
            initial_teeth_num = 18

        elif gear.pressure_angle == 25:
            initial_teeth_num = 13

        # elif gear.pressure_angle == 14.5:
        #     initial_teeth_num = 32
        else:
            raise ValueError(f"pressure_angle={gear.pressure_angle} "
                             f"degrees but it can only be 20/25 degrees")

        modulus_list = [0.3, 0.4, 0.5, 0.8, 1, 1.25, 1.5, 2, 2.5, 3, 4, 5, 6, 8, 10, 12, 16, 20, 25]
        modulus_list.sort(reverse=True)

        results_list = []
        for modulus in modulus_list:
            # set gear modulus
            gear.modulus = modulus
            gear.teeth_num = initial_teeth_num  # updating attribute

            # because the width range is (3πm,5πm) the initial guess is 4πm
            initial_width = 4 * pi * transmission.gear_ratio
            gear.width = initial_width

            # print(modulus, initial_teeth_num, initial_width)
            while True:
                # update Yj factors after attribute changed
                transmission.gear1.Y_j(transmission.gear1, transmission.gear2)

                # KH is a function of the gear width,
                # we assumed an initial width of 4πm to calculate KH,
                # we recalculate KH and the width until KH converges
                kh_not_converging = False
                while True:
                    # calculate minimum gear width for bending and contact stress
                    bending_minimum_width = transmission.minimum_width_for_bending(gear)
                    contact_minimum_width = transmission.minimum_width_for_contact(gear)

                    # the new width is the max value of the two minimum width
                    new_width = max(bending_minimum_width, contact_minimum_width)
                    if new_width > 1020:
                        print("error: KH is not converging for m=", modulus)
                        kh_not_converging = True
                        break

                    # save old KH for comparison later
                    oldKH = gear.KH

                    # save new gear width
                    gear.width = new_width

                    # update Yj factors after width changed
                    transmission.gear1.Y_j(transmission.gear1, transmission.gear2)

                    newKH = gear.KH
                    # checking convergence
                    if abs(newKH - oldKH) < 1e-6:
                        break

                # in case KH couldn't converge exit optimization for current modulus
                if kh_not_converging:
                    break

                alpha = contact_minimum_width / bending_minimum_width
                centers_distance = 0.5 * gear.modulus * gear.teeth_num * (
                        transmission.gear_ratio + 1)
                volume = 0.25 * pi * (gear.pitch_diameter ** 2) * gear.width
                f_string = f"m={gear.modulus}, N={gear.teeth_num}, b={gear.width:.2f}," \
                           f"spring_index={centers_distance:.2f}, V={volume:.2f}, α={alpha:.4f}"

                if gear.width < 3 * pi * gear.modulus:
                    # gear width is less than 3πm (b<3πm), gear width should be increased
                    # because the initial teeth number is minimal decrease modulus

                    modulus_list = [0.3, 0.4, 0.5, 0.8, 1, 1.25, 1.5, 2, 2.5, 3, 4, 5, 6, 8, 10, 12,
                                    16, 20, 25]
                    # get previous modulus on the list
                    if gear.modulus > 0.3:
                        new_modulus = modulus_list[modulus_list.index(gear.modulus) - 1]
                    else:
                        raise ValueError(
                            "at Optimize: b<3πm but the modulus is the lowest possible")

                    if verbose:
                        # printing step result
                        v = gear.tangent_velocity
                        v_max = gear.maximum_velocity
                        msg = "b<3πm" if v < v_max else "b<3πm, v>v_max"
                        print(f_string, ',', msg)

                    # setting changes
                    gear.modulus = new_modulus

                elif gear.width > 5 * pi * gear.modulus:
                    # gear width is more than 5πm (b>5πm), gear width should be decreased
                    # because initial modulus is maximal increase teeth number

                    if verbose:
                        # print step result
                        v = gear.tangent_velocity
                        v_max = gear.maximum_velocity
                        msg = "b>5πm" if v < v_max else "b>5πm, v>v_max"
                        print(f_string, ',', msg)

                    # increasing teeth number by one
                    # (note: the gear teeth number can't exceed the
                    # biggest number specified in the Yj factor tables
                    # and for a pinion a number that will make its gear pass this number.
                    # this error is handled at the Yj function)
                    gear.teeth_num += 1
                else:
                    # gear width is within range (3πm<b<5πm)
                    if alpha > 1:
                        # if α>1 increase number of teeth
                        if verbose:
                            # print step result
                            v = gear.tangent_velocity
                            v_max = gear.maximum_velocity
                            print(f"{f_string}, 3πm<b<5πm, α>1" if v < v_max
                                  else "3πm<b<5πm, α>1, v>v_max")

                        # add result to least of viable results
                        results_list.append(
                            {'m': gear.modulus, 'N': gear.teeth_num, 'b': gear.width,
                             'spring_index': centers_distance, 'V': volume, 'alpha': alpha})
                        # increase teeth number
                        gear.teeth_num += 1
                    else:
                        # if α<=1 stop optimization and return the optimized value
                        # and a list of all viable options

                        # print optimization progress
                        if verbose:
                            # print step result
                            v = gear.tangent_velocity
                            v_max = gear.maximum_velocity
                            print(f"{f_string}, 3πm<b<5πm, α<=1" if v < v_max
                                  else "3πm<b<5πm, α<=1, v>v_max")

                        # find optimize feature
                        # optimize by width
                        if optimize_feature == 'width':
                            # create list of widths
                            width_list = [(dic['b'], index) for index, dic in
                                          enumerate(results_list)]
                            # get minimum width index
                            result_index = min(width_list)[1]
                            optimized_result = results_list[result_index]

                        # optimize by volume
                        elif optimize_feature == 'volume':
                            # create list of volumes
                            volume_list = [(dic['V'], index) for index, dic in
                                           enumerate(results_list)]
                            # get minimum volume index
                            result_index = min(volume_list)[1]
                            optimized_result = results_list[result_index]

                        # optimize by center distance
                        elif optimize_feature == 'center':
                            # create list of center distances
                            center_list = [(dic['spring_index'], index) for index, dic in
                                           enumerate(results_list)]
                            # get minimum center distance index
                            result_index = min(center_list)[1]
                            optimized_result = results_list[result_index]

                        elif optimize_feature == 'all':
                            width_list = [(dic['b'], index) for index, dic in
                                          enumerate(results_list)]
                            volume_list = [(dic['V'], index) for index, dic in
                                           enumerate(results_list)]
                            center_list = [(dic['spring_index'], index) for index, dic in
                                           enumerate(results_list)]

                            optimized_result = {'optimized width': results_list[min(width_list)[1]],
                                                'optimized volume': results_list[
                                                    min(volume_list)[1]],
                                                'optimized center': results_list[
                                                    min(center_list)[1]]}
                        else:
                            raise Exception(f"optimize_feature={optimize_feature} is invalid")

                        # restore gear parameters
                        gear.teeth_num = original_teeth_num
                        gear.width = original_width
                        gear.modulus = original_modulus

                        return optimized_result, results_list

    def calc_centers_distance(self, gear_ratio):
        """Calculate the distance between the centers of the gears
        :param float gear_ratio: transmissions gear ratio

        :return: the distance between the transmissions gears centers  in [mm]
        :rtype: float
        """
        centers_distance = 0.5 * self.modulus * self.teeth_num * (gear_ratio + 1)
        return centers_distance

    @staticmethod
    def create_new_gear(gear2_prop):
        """Return new instance of SpurGear

        :param dict gear2_prop: gear properties

        :returns: SpurGear object
        :type: SpurGear
        """
        return SpurGear(**gear2_prop)

    @staticmethod
    def format_properties(properties):
        """Remove excess attributes form properties

        :param list properties: list of gear properties

        :returns: A dictionary of properties
        :rtype: dict
        """
        prop_list = ['modulus', 'teeth_num', 'rpm', 'Qv', 'width', 'bearing_span',
                     'pinion_offset', 'enclosure',
                     'hardness', 'work_hours', 'number_of_cycles', 'pressure_angle', 'grade',
                     'crowned', 'adjusted',
                     'sensitive_use', 'nitriding', 'case_carb', 'material']
        # remove contact ratio from dictionary so it won't interfere with the new gear instantiation
        new_properties = {key: properties[key] for key in properties if key in prop_list}
        return new_properties

    def check_compatibility(self, gear):
        """Raise error if the pressure angle, helix angle or modulus of the gears don't match

        :param HelicalGear gear: gear object

        :raises ValueError("gear1 and gear2 have mismatch pressure_angle"):
        :raises ValueError("gear1 and gear2 have mismatch Helix_angle"):
        :raises ValueError("gear1 and gear2 are not the same type, they are no compatible"):
        :raises ValueError("gear1 and gear2 have mismatch modulus"):
        """
        if self.pressure_angle != gear.pressure_angle:
            # if pressure angles are different raise error
            raise ValueError("gear1 and gear2 have mismatch pressure_angle")

        if self.modulus != gear.modulus:
            # if modulus are different raise error
            raise ValueError("gear1 and gear2 have mismatch modulus")

        if type(self) is not type(gear):
            raise ValueError("gear1 and gear2 are not the same type, they are no compatible")



from unittest import TestCase
from math import pi, sqrt

from me_toolbox.fasteners import Bolt


class TestBolt(TestCase):

    def setUp(self):
        diameter = 10
        pitch = 1.5
        length = 20
        threaded_length = 15
        grade = '12.9'
        E = 207e3

        Sp, Sut, Sy = Bolt.get_strength_prop(diameter, grade)
        self.bolt = Bolt(diameter, pitch, length, threaded_length, Sy, Sut, Sp, E)

    def test_height(self):
        self.assertAlmostEqual(self.bolt.height, self.bolt.pitch * sqrt(3) / 2)

    def test_mean_diam(self):
        self.assertAlmostEqual(self.bolt.mean_diameter,
                               self.bolt.diameter - (5 / 8) * self.bolt.height)

    def test_root_diam(self):
        self.assertAlmostEqual(self.bolt.minor_diameter,
                               self.bolt.diameter - (5 / 4) * self.bolt.height)

    def test_pitch_diam(self):
        self.assertAlmostEqual(self.bolt.pitch_diameter,
                               self.bolt.diameter - (3 / 8) * self.bolt.height)

    def test_head_diam(self):
        self.assertAlmostEqual(self.bolt.head_diameter, 1.5 * self.bolt.diameter)

    def test_unthreaded_length(self):
        self.assertAlmostEqual(self.bolt.shank_length,
                               self.bolt.length - self.bolt.thread_length)

    def test_nominal_area(self):
        self.assertAlmostEqual(self.bolt.nominal_area,
                               0.25 * pi * self.bolt.diameter ** 2)

    def test_thread_length(self):
        self.assertAlmostEqual(self.bolt.thread_length, 26)

    def test_stress_area(self):
        self.assertAlmostEqual(self.bolt.stress_area, 57.9895969018452)

    def test_minor_area(self):
        self.assertAlmostEqual(self.bolt.minor_area, 52.29231784971083)



"""Module containing the HelicalGear class"""
# I want the variables names to be the same as in AGMA pylint: disable=invalid-name
from math import sin, cos, radians, pi, tan, atan, sqrt, degrees
import os
import numpy as np
from me_toolbox.gears import SpurGear  # for inheritance
from me_toolbox.tools import table_interpolation


class HelicalGear(SpurGear):
    """A gear object that contains all of its design parameters (AGMA 2001-D04)"""
    def __repr__(self):
        return f"{self.__class__.__name__}(m={self.modulus}, N={self.teeth_num}, " \
               f"\N{GREEK SMALL LETTER PHI}={self.pressure_angle}, " \
               f"\N{GREEK SMALL LETTER PSI}={self.helix_angle}, b={self.width})"

    def __init__(self, modulus, teeth_num, rpm, Qv, width, bearing_span, pinion_offset,
                 enclosure, hardness, pressure_angle, helix_angle, grade, work_hours=0,
                 number_of_cycles=0, crowned=False, adjusted=False, sensitive_use=False,
                 nitriding=False, case_carb=False, material='steel'):
        """Instantiating HelicalGear object which inherits from SpurGear

        :param float helix_angle: The gear helix angle (20 or 25)

        :returns: HelicalGear object
        :rtype: HelicalGear
        """
        super().__init__(modulus, teeth_num, rpm, Qv, width, bearing_span, pinion_offset,
                         enclosure, hardness, pressure_angle, grade, work_hours, number_of_cycles,
                         crowned, adjusted, sensitive_use, nitriding, case_carb, material)

        self.helix_angle = helix_angle

    @property
    def pitch_diameter(self):
        """Calculate pitch diameter

        :returns: Gear's pitch diameter
        :rtype: float
        """
        return self.teeth_num * self.tangent_modulus  # pitch diameter [mm]

    @property
    def tangent_pitch(self):
        """Calculate tangent pitch

        :returns: Gear's tangent pitch
        :rtype: float
        """

        return self.pitch / cos(radians(self.helix_angle))

    @property
    def axial_pitch(self):
        """Calculate axial pitch

        :returns: Gear's axial pitch
        :rtype: float
        """
        return self.pitch / sin(radians(self.helix_angle))

    @property
    def tangent_modulus(self):
        """Calculate tangent modulus

        :returns: Gear's tangent modulus
        :rtype: float
        """
        return self.modulus / cos(radians(self.helix_angle))

    @property
    def axial_modulus(self):
        """Calculate axial modulus

        :returns: Gear's axial modulus
        :rtype: float
        """
        return self.modulus / sin(radians(self.helix_angle))

    @property
    def tangent_pressure_angle(self):
        """Calculate tangent pressure angle

        :returns: Gear's tangent pressure angle
        :rtype: float
        """
        return degrees(atan(tan(radians(self.pressure_angle)) / cos(radians(self.helix_angle))))

    @staticmethod
    def Y_j(gear1, gear2):
        """Return Geometry factors for helical gear and pinion
        Yj is dependent on the gear ratio, helix angle and both gears teeth numbers

        :param HelicalGear gear1: Helical gear object
        :param HelicalGear gear2: Helical gear object

        :return: Yj - Geometric factor
        :rtype: float
        """

        Np = gear1.teeth_num
        Ng = gear2.teeth_num
        helix_angle = gear1.helix_angle

        # files path
        j75_path = os.path.dirname(__file__) + "\\tables\\J75 - helix gear geometry factors.csv"
        jPrime_path = os.path.dirname(
            __file__) + "\\tables\\JPrime - helix gear geometry factors.csv"

        # load data
        j75_data = np.genfromtxt(j75_path, delimiter=',')
        jPrime_data = np.genfromtxt(jPrime_path, delimiter=',')

        # data interpolation
        j75 = table_interpolation(Np, helix_angle, j75_data)
        jPrime = table_interpolation(Ng, helix_angle, jPrime_data)

        # calculate geometric factor Yj
        Y_j = j75 * jPrime

        gear1.Yj = Y_j
        gear2.Yj = Y_j

    @staticmethod
    def ZI(gear1, gear2):
        """Return geometric factor for contact failure
        ZI is dependent on the gear ratio, pressure angle, modulus and both gears teeth numbers

        :param HelicalGear gear1: gear object
        :param HelicalGear gear2: gear object

        :return: ZI - geometric factor
        :rtype: float
        """

        mG = gear2.teeth_num / gear1.teeth_num
        phi_t = radians(gear1.tangent_pressure_angle)
        phi_n = radians(gear1.pressure_angle)
        modulus = gear1.modulus
        tan_mod = gear1.tangent_modulus
        Np = gear1.teeth_num
        Ng = gear2.teeth_num
        z1 = sqrt((0.5 * tan_mod * Np + modulus) ** 2 - (0.5 * tan_mod * Np * cos(phi_t)) ** 2)
        z2 = sqrt((0.5 * tan_mod * Ng + modulus) ** 2 - (0.5 * tan_mod * Ng * cos(phi_t)) ** 2)
        z3 = 0.5 * tan_mod * Np * (1 + mG) * sin(phi_t)

        if z1 > z3 and z2 > z3:
            z = z3
        elif z1 > z3:
            z = z2
        elif z2 > z3:
            z = z1
        else:
            z = z1 + z2 - z3
        if gear1.width >= 2 * gear1.axial_pitch:
            mN = (gear1.pitch * cos(phi_n)) / (0.95 * z)
        else:
            raise ValueError(f"@ ZI factor: gear width should be at least two times the axial pitch"
                             f"2Px={2 * gear1.axial_pitch:.2f}")
        Z_I = (cos(phi_t) * sin(phi_t) * mG) / (2 * mN * (mG + 1))
        # print(f"ZI={Z_I}, mG={mG}, modulus={modulus}, Np={Np}, Ng={Ng}, Px={gear1.axial_pitch}")
        return Z_I

    @staticmethod
    def calc_forces(gear, power):
        """Calculate forces on helical gear

        :param HelicalGear gear: gear object
        :param float power: power

        :returns: Wt - tangent max_force, Wr - radial max_force, Wx - axial max_force in [N]
        :rtype: tuple[float, float, float]
        """

        phi_t = radians(gear.tangent_pressure_angle)
        Wt = (60e3 / pi) * (power / (gear.pitch_diameter * gear.rpm))
        Wr = Wt * tan(radians(phi_t))
        Wx = Wt * tan(radians(gear.helix_angle))
        return Wt, Wr, Wx

    def optimization(self, transmission, optimize_feature='all', verbose=False):
        """Perform gear optimization

        :param gears.transmission.Transmission transmission: Transmission object
            associated with the gear
        :param str optimize_feature: property to optimize for ('width'/'volume'/'center')
        :param bool verbose: print optimization stages

        :return: optimized result (width in mm, volume in mm^3, center distance in mm)
        :rtype: dict
        """
        gear = self

        # saving original attribute values
        original_teeth_num = gear.teeth_num  # saving original attribute
        original_width = gear.width
        original_modulus = gear.modulus

        # choosing the minimum number of teeth to start with
        initial_teeth_num = 21

        modulus_list = [0.3, 0.4, 0.5, 0.8, 1, 1.25, 1.5, 2, 2.5, 3, 4, 5, 6, 8, 10, 12, 16, 20, 25]
        modulus_list.sort(reverse=True)

        results_list = []
        for modulus in modulus_list:
            # set gear modulus
            gear.modulus = modulus
            gear.teeth_num = initial_teeth_num  # updating attribute

            # because the width range is (3πm,5πm) the initial guess is 4πm
            # initial_width = 0.5 * (2 * self.axial_pitch + self.pitch_diameter)
            initial_width = gear.axial_pitch + 0.5 * gear.pitch_diameter
            gear.width = initial_width
            # print(modulus, initial_teeth_num, initial_width)
            while True:
                # update Yj factors after attribute changed
                transmission.gear1.Y_j(transmission.gear1, transmission.gear2)
                try:
                    transmission.ZI = self.ZI(transmission.gear1, transmission.gear2)
                except ValueError:
                    pass

                # KH is function of the gear width,
                # we assumed an initial width of 4πm to calculate KH,
                # we recalculate KH and the width until KH converges
                kh_not_converging = False
                while True:

                    # calculate minimum gear width for bending and contact stress
                    bending_minimum_width = transmission.minimum_width_for_bending(gear)
                    contact_minimum_width = transmission.minimum_width_for_contact(gear)
                    # print(f"bending={bending_minimum_width}, contact={contact_minimum_width}")

                    # the new width is the max value of the two minimum width
                    new_width = max(bending_minimum_width, contact_minimum_width)
                    if new_width > 1020:
                        print("error: KH is not converging for m=", self.modulus)
                        kh_not_converging = True
                        break

                    # save old KH for comparison later
                    oldKH = gear.KH

                    # save new gear width
                    gear.width = new_width

                    # update Yj factors after width changed
                    transmission.gear1.Y_j(transmission.gear1, transmission.gear2)
                    try:
                        transmission.ZI = self.ZI(transmission.gear1, transmission.gear2)
                    except ValueError:
                        pass

                    newKH = gear.KH
                    # checking convergence
                    if abs(newKH - oldKH) < 1e-6:
                        break

                # in case KH couldn't converge exit optimization for current modulus
                if kh_not_converging:
                    break

                alpha = contact_minimum_width / bending_minimum_width
                mG = transmission.gear_ratio
                centers_distance = 0.5 * gear.modulus * gear.teeth_num * (mG + 1)
                volume = 0.25 * pi * (gear.pitch_diameter ** 2) * gear.width
                f_string = f"m={gear.modulus}, N={gear.teeth_num}, b={gear.width:.2f}," \
                           f"spring_index={centers_distance:.2f}, V={volume:.2f}, α={alpha:.4f}"

                if gear.width < 2 * pi * gear.axial_pitch:
                    # gear width is less than 2Px (b<Px), gear width should be increased
                    # because the initial teeth number is minimal decrease modulus

                    modulus_list = [0.3, 0.4, 0.5, 0.8, 1, 1.25, 1.5, 2, 2.5, 3, 4, 5, 6, 8, 10, 12,
                                    16, 20, 25]
                    # get previous modulus on the list
                    if gear.modulus > 0.3:
                        new_modulus = modulus_list[modulus_list.index(gear.modulus) - 1]
                    else:
                        raise ValueError("@ Optimize: b<2Px but the modulus is the lowest possible")

                    if verbose:
                        # printing step result
                        v = gear.tangent_velocity
                        v_max = gear.maximum_velocity
                        if not isinstance(v_max, str):
                            msg = "b<2Px" if v < v_max else "b<2Px, v>v_max"
                        else:
                            msg = "b<2Px"
                        print(f_string, ',', msg)

                    # setting changes
                    gear.modulus = new_modulus

                elif gear.width > gear.pitch_diameter:
                    # gear width is more than 5πm (b>Pd), gear width should be decreased
                    # because initial modulus is maximal increase teeth number

                    if verbose:
                        # print step result
                        v = gear.tangent_velocity
                        v_max = gear.maximum_velocity
                        if not isinstance(v_max, str):
                            msg = "b>Pd" if v < v_max else "b>Pd, v>v_max"
                        else:
                            msg = "b>Pd"
                        print(f_string, ',', msg)

                    # increasing teeth number by one
                    # (note: the gear teeth number can't exceed the biggest number specified
                    # in the Yj factor tables
                    # and for pinion a number that will make its gear pass this number.
                    # this error is handled at the Yj function)
                    gear.teeth_num += 1
                else:
                    # gear width is within range (3πm<b<5πm)
                    if alpha > 1:
                        # if α>1 increase number of teeth
                        if verbose:
                            # print step result
                            v = gear.tangent_velocity
                            v_max = gear.maximum_velocity
                            if not isinstance(v_max, str):
                                print(f"{f_string}, 2Px<b<Pd, α>1" if v < v_max
                                      else "2Px<b<Pd, α>1, v>v_max")
                            else:
                                print(f"{f_string}, 2Px<b<Pd, α>1")
                        # add result to least of viable results
                        results_list.append({'m': gear.modulus, 'N': gear.teeth_num,
                                             'b': gear.width, 'spring_index': centers_distance,
                                             'V': volume, 'alpha': alpha})
                        # increase teeth number
                        gear.teeth_num += 1
                    else:
                        # if α<=1 stop optimization and return the optimized value
                        # and a list of all viable options

                        # print optimization progress
                        if verbose:
                            # print step result
                            v = gear.tangent_velocity
                            v_max = gear.maximum_velocity
                            if not isinstance(v_max, str):
                                print(f"{f_string}, 2Px<b<Pd, α<=1" if v < v_max
                                      else "2Px<b<Pd, α<=1, v>v_max")
                            else:
                                print(f"{f_string}, 2Px<b<Pd, α<=1")

                        # find optimize feature
                        # optimize by width
                        if optimize_feature == 'width':
                            # create list of widths
                            width_list = [(dic['b'], index) for index, dic in
                                          enumerate(results_list)]
                            # get minimum width index
                            result_index = min(width_list)[1]
                            opt_res = results_list[result_index]

                        # optimize by volume
                        elif optimize_feature == 'volume':
                            # create list of volumes
                            volume_list = [(dic['V'], index) for index, dic in
                                           enumerate(results_list)]
                            # get minimum volume index
                            result_index = min(volume_list)[1]
                            opt_res = results_list[result_index]

                        # optimize by center distance
                        elif optimize_feature == 'center':
                            # create list of center distances
                            center_list = [(dic['spring_index'], index) for index, dic in
                                           enumerate(results_list)]
                            # get minimum center distance index
                            result_index = min(center_list)[1]
                            opt_res = results_list[result_index]

                        elif optimize_feature == 'all':
                            width_list = [(dic['b'], index) for index, dic in
                                          enumerate(results_list)]
                            volume_list = [(dic['V'], index) for index, dic in
                                           enumerate(results_list)]
                            center_list = [(dic['spring_index'], index) for index, dic in
                                           enumerate(results_list)]

                            opt_res = {'optimized width': results_list[min(width_list)[1]],
                                       'optimized volume': results_list[min(volume_list)[1]],
                                       'optimized center': results_list[min(center_list)[1]]}
                        else:
                            raise Exception(f"optimize_feature={optimize_feature} is invalid")

                        # restore gear parameters
                        gear.teeth_num = original_teeth_num
                        gear.width = original_width
                        gear.modulus = original_modulus

                        return opt_res, results_list

    def calc_centers_distance(self, gear_ratio):
        """ calculate the distance between the centers of the gears

        :param float gear_ratio: transmissions gear ratio

        :return: the distance between the transmissions gears centers in [mm]
        :rtype: float
        """
        centers_distance = 0.5 * self.tangent_modulus * self.teeth_num * (gear_ratio + 1)
        return centers_distance

    @staticmethod
    def create_new_gear(gear2_prop):
        """ return new instance of HelicalGear

        :param dict gear2_prop: gear properties

        :returns: HelicalGear object
        :type: HelicalGear
        """
        return HelicalGear(**gear2_prop)

    @staticmethod
    def format_properties(properties):
        """Remove excess attributes form properties

        :param list properties: list of gear properties

        :returns: A dictionary of properties
        :rtype: dict
        """
        prop_list = ['name', 'modulus', 'teeth_num', 'rpm', 'Qv', 'width', 'bearing_span',
                     'pinion_offset', 'enclosure', 'hardness', 'work_hours', 'number_of_cycles',
                     'pressure_angle', 'grade', 'crowned', 'adjusted', 'sensitive_use', 'nitriding',
                     'case_carb', 'material', 'helix_angle']
        # remove contact ratio from dictionary so it won't interfere with the new gear instantiation
        new_properties = {key: properties[key] for key in properties if key in prop_list}
        return new_properties

    def check_compatibility(self, gear):
        """Raise error if the pressure angle, helix angle or modulus of the gears don't match

        :param HelicalGear gear: gear object

        :raises ValueError("gear1 and gear2 have mismatch pressure_angle"):
        :raises ValueError("gear1 and gear2 have mismatch Helix_angle"):
        :raises ValueError("gear1 and gear2 are not the same type, they are no compatible"):
        :raises ValueError("gear1 and gear2 have mismatch modulus"):
        """

        # check pressure angles
        if self.pressure_angle != gear.pressure_angle:
            # if pressure angles are different raise error
            raise ValueError("gear1 and gear2 have mismatch pressure_angle")

        # check helix angles
        try:
            if self.helix_angle != gear.helix_angle:
                # if pressure angles are different raise error
                raise ValueError("gear1 and gear2 have mismatch Helix_angle")
        except AttributeError as err:
            raise ValueError("gear1 and gear2 are not the same type, they are no compatible") \
                from err

        # check modulus
        if self.modulus != gear.modulus:
            # if modulus are different raise error
            raise ValueError("gear1 and gear2 have mismatch modulus")



"""A module containing stress calculation functions"""


# TODO: better stress calculation
def uniform_stress(F, A):
    """Returns stress assuming uniform distribution in cross section

    :param float F: Force
    :param float A: Cross section area

    :returns Stress
    :rtype: float
    """
    return F / A


def bending_stress(My, Iy, z, Mz=None, Iz=None, y=None):
    """Bending stress in principle system

    :param float My: Moment in the y direction
    :param float Mz: Moment in the y direction
    :param float Iy: Area moment of inertia for the y direction
    :param float Iz: Area moment of inertia for the zeta direction
    :param float z: Coordinate
    :param float y: Coordinate

    :returns: Bending Stress in cross section
    :rtype: float
    """

    if Mz is None:
        return (My / Iy) * z
    else:
        return (My / Iy) * z, -(Mz / Iz) * y


def shear_bending_stress(V, Q, I, b):
    """ Shear stresses due to bending

    :param float V: Shear max_force in y direction
    :param float Q: first moment of in cross section in y direction
    :param float I: Area moment of inertia around the y axis
    :param float b: thickness

    :returns: Shear stress resulting from bending
    :rtype: float
    """

    return (V * Q) / (I * b)


def torsion_stress(T, r, J):
    """ Torsion stresses

    :param float T: Cross section torque
    :param float r: Stress radius coordinate
    :param float J: Cross section polar moment of inertia

    :returns: Torsion Stress at r
    :rtype: float
    """
    return (T * r) / J


def max_shear_stress(V, A, shape):
    """Returns The maximum shear stress for known shapes

    :param float V: Shear Stress
    :param float A: Cross section area
    :param str shape: Cross section shape

    :returns: Maximum shear Stress
    :rtype: float
    """
    if shape == 'circle':
        return (4 * V) / (3 * A)
    elif shape == ' rectangle':
        return (3 * V) / (2 * A)
    else:
        raise ValueError(f"shape = {shape} is unknown")



"""module containing misc functions"""
import numpy as np


def print_atributes(obj):
    """prints all of the attributes of an object
    excluding the dunder attributes and attributes
    with single leading underscore

    :param obj: any object
    """
    # attribute_list = [attribute for attribute in dir(obj) if
    #                   not callable(getattr(obj, attribute)) and attribute.startswith(
    #                       '__') is False and attribute.startswith('_') is False]

    attribute_list = []

    for attribute in dir(obj):
        try:
            if (not callable(getattr(obj, attribute)) and attribute.startswith('__') is False and
                    attribute.startswith('_') is False):
                attribute_list.append(attribute)
        except Exception:
            continue

    for item in attribute_list:
        print(f"{item} : {getattr(obj, item)}")


def parse_input(obj, fields, kwargs):
    """adds attributes specified in kwargs to obj only if they are specified in fields
    fields contain dictionary of the anticipated attributes and their default value
    if one exists, if not use the '' as place holder

    :param obj: An object to which add the attributes
    :param dict fields: A dictionary containing the desired attributes with default values
    :param dict kwargs: dictionary containing the keywords to check

    .. code-block:: python

        from tools import parse_input
        from tools import print_atributes
        class P:
            def __init__(self,**kwargs):
                fields = {'name':'', 'age':'30', 'fav_color':'', 'hobby':'sky'}
                parse_input(self, fields, kwargs)

        p=P(name='Omri', fav_color='blue')
        print_atributes(p)
    """
    for key, val in kwargs.items():
        if key in fields:
            setattr(obj, key, val)
        else:
            raise TypeError(f"__init__() got an unexpected keyword argument '{key}'")

    # check if all attributes were assigned

    error_list = []
    for field in fields:
        if field not in kwargs:  # check if a field was not filled
            if fields[field] != '':  # check if a default value exists
                setattr(obj, field, fields[field])
            else:
                error_list.append(field)

    # raise error if a field is not satisfied
    if len(error_list) == 1:
        raise TypeError(f"__init__() missing 1 required keyword argument: '{error_list[0]}'")
    elif len(error_list) > 1:
        msg = f"__init__() missing {len(error_list)} required keyword arguments: "
        raise TypeError(msg + f"{str(error_list[:-1]).strip('[]')} and {str(error_list[-1])}")


# unit conversions
def pound_force_to_newtons(val):
    """convert pound max_force to newtons

    :param float or list val: value to convert

    :returns: converted value
    :rtype: float or tuple
    """
    try:
        return val * 4.448
    except TypeError:
        return [x * 4.448 for x in val]


def newtons_to_pound_force(val):
    """convert newtons to pound max_force

    :param float or list val: value to convert

    :returns: converted value
    :rtype: float or tuple
    """
    try:
        return val / 4.448
    except TypeError:
        return [x / 4.448 for x in val]


def inch_to_millimetre(val):
    """convert inches in to millimeters

    :param float or list val: value to convert

    :returns: converted value
    :rtype: float or tuple
    """
    try:
        return val * 25.4
    except TypeError:
        return [x * 25.4 for x in val]


def millimetre_to_inch(val):
    """convert millimetres to inches

    :param float or list val: value to convert

    :returns: converted value
    :rtype: float or tuple
    """
    try:
        return val / 25.4
    except TypeError:
        return [x / 25.4 for x in val]


def percent_to_decimal(var):
    """if input in percent (>=1) convert to decimal"""
    if not isinstance(var, list) and not isinstance(var, tuple):
        dec = var / 100 if var >= 1 else var
    else:
        dec = []
        for val in var:
            if val >= 1:
                # if var is in percentage form divide by 100
                dec.append(val / 100)
            else:
                # if var is in decimal form no correction needed
                dec.append(val)
    return dec


def pol2cart(rho, phi):
    x = rho * np.cos(np.deg2rad(phi))
    y = rho * np.sin(np.deg2rad(phi))
    return x, y



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="me_toolbox",
    version="0.0.13",
    author="Omri Stein",
    author_email="omri.stein@gmail.com",
    description="Mechanical engineering design tools",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/OmriStein/MEtoolbox",
    project_urls={
        "Bug Tracker": "https://github.com/OmriStein/MEtoolbox/issues",
    },
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Education',
        'Topic :: Scientific/Engineering',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    package_dir={"me_toolbox": "me_toolbox"},
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=['sympy', 'numpy', 'mpmath'],
    python_requires=">=3.9",
)



from me_toolbox.springs.spring import Spring
from me_toolbox.springs.helical_compression_spring import HelicalCompressionSpring
from me_toolbox.springs.extension_spring import ExtensionSpring
from me_toolbox.springs.helical_torsion_spring import HelicalTorsionSpring



"""module containing the bolts base class Bolt"""
from collections import namedtuple
from math import sqrt, pi
from mpmath import sec

from me_toolbox.tools import print_atributes
from me_toolbox.fatigue import EnduranceLimit


class Bolt:
    """Bolt class containing basic geometry attributes"""
    angle = 60

    def __repr__(self):
        return f"Bolt(diameter={self.diameter}, pitch={self.pitch}, length={self.length}, " \
               f"thread_length={self.thread_length}, yield_strength={self.yield_strength}, " \
               f"tensile_strength={self.tensile_strength}, proof_strength={self.proof_strength}, " \
               f"elastic_modulus={self.elastic_modulus}, Kf={self.Kf})"

    def __str__(self):
        return f"Bolt({self.diameter}x{self.pitch}x{self.length})"

    def __init__(self, diameter, pitch, length, thread_length,
                 yield_strength, tensile_strength, proof_strength, elastic_modulus, Kf=None):
        """Initialise Bolt object
        :param float diameter: Nominal diameter
        :param float pitch: Thread's pitch
        :param float length: Bolt's length
        :param float thread_length: Bolt's length
        :param float yield_strength: Bolt's yield strength
        :param float tensile_strength: Bolt's tensile strength
        :param float proof_strength: Bolt's proof strength (if unknown use 85% of yield strength)
        :param float elastic_modulus: Bolt's elastic modulus
        :param float Kf: Dynamic stress concentration factor
        """
        self.diameter = diameter
        self.pitch = pitch
        self.length = length
        self.thread_length = thread_length

        self.yield_strength = yield_strength
        self.tensile_strength = tensile_strength
        self.proof_strength = proof_strength
        self.elastic_modulus = elastic_modulus
        self.Kf = Kf

    def get_info(self):
        """print all the bolt's properties"""
        print_atributes(self)

    @property
    def height(self):
        """Height of fundamental triangle (H)"""
        return self.pitch * sqrt(3) / 2

    @property
    def minor_diameter(self):
        """Minor diameter (dr) as calculated in Table 8-1 of Shigley"""
        return self.diameter - 1.226869 * self.pitch

    @property
    def mean_diameter(self):
        """Mean diameter(dm) of the nominal and root diameters(dr)"""
        # return self.diameter - (5 / 8) * self.height
        return 0.5 * (self.diameter + self.minor_diameter)

    @property
    def pitch_diameter(self):
        """Pitch diameter(dp) as calculated in Table 8-1 of Shigley"""
        return self.diameter - 0.649519 * self.pitch

    @property
    def head_diameter(self):
        """Estimated diameter of bolt head (D)"""
        return 1.5 * self.diameter

    @property
    def shank_length(self):
        """The unthreaded section of the bolt (ld)"""
        return self.length - self.thread_length

    @property
    def nominal_area(self):
        """Area of the bolt's nominal diameter (Ad)"""
        return 0.25 * pi * self.diameter ** 2

    @property
    def minor_area(self):
        """Area of the bolt's root or minor diameter (Ar)"""
        return 0.25 * pi * self.minor_diameter ** 2

    @property
    def proof_load(self):
        """Proof load of the bolt (Fp)"""
        return self.proof_strength * self.stress_area

    @property
    def stress_area(self):
        """Tensile stress area (At)"""
        dt = 0.5 * (self.minor_diameter + self.pitch_diameter)
        return 0.25 * pi * dt ** 2

    def estimate_preload(self, reused):
        """Estimated Pre-Load(Fi) for both static and fatigue loading
                :param bool reused: True for reused fastener or False for a permanent joint
                """
        return 0.75 * self.proof_load if reused else 0.90 * self.proof_load

    def preload2torque(self, preload, thread_friction, collar_friction):
        """calculating tightening torque from preload"""
        Fi, d, dm = preload, self.diameter, self.mean_diameter
        length = self.pitch  # for single start
        tanG = length / (pi * dm)
        alpha = self.angle
        return Fi * d * ((dm / (2 * d)) *
                         ((tanG + thread_friction * sec(alpha)) /
                          (1 - thread_friction * tanG * sec(alpha))) + 0.625*collar_friction)

    def torque2preload(self, torque):
        pass

    def endurance_limit(self, unmodified_Se, surface_finish, temp, reliability):
        """Calls the EnduranceLimit class with the right parameters for a bolt
        :param float unmodified_Se: Unmodified endurance strength
        :param string surface_finish: the bolt's surface finish
            options:['ground', 'machined', 'cold-drawn', 'hot-rolled', 'as forged']
        :param float temp: The bolt's operating temperature in deg C
        :param float reliability: bolt's reliability

        :returns: An endurance limit object
        :rtype: EnduranceLimit
        """
        finishes = ['ground', 'machined', 'cold-drawn', 'hot-rolled', 'as forged']

        if surface_finish not in finishes:
            raise Exception(f"The surface finish({surface_finish}) is not one of the acceptable"
                            f"surface finishes[{finishes}]")

        Se = EnduranceLimit(unmodified_Se, Sut=self.tensile_strength, surface_finish=surface_finish,
                            rotating=False, max_normal_stress=1, max_bending_stress=0,
                            stress_type='multiple', temp=temp,
                            reliability=reliability,
                            diameter=sqrt((4*self.stress_area)/pi))
        return Se

    @staticmethod
    def get_strength_prop(diameter, grade):
        """Returns the proof strength, tensile strength and yield strength
            of the bolt in accordance with its grade
            :param float diameter: Bolt's diameter
            :param string grade: Bolt's grade or class
            options: ['4.6', '4.8', '5.8', '8.8', '9.8', '10.9', '12.9']

            :returns: Proof-Strength, Ultimate-tensile-strength, Yield-strength
            :rtype: list
            """
        grade_prop = namedtuple('Grade', ['low', 'high', 'Sp', 'Sut', 'Sy'])
        grade_list = {'4.6': grade_prop(5, 36, 225, 400, 240),
                      '4.8': grade_prop(1.6, 16, 310, 420, 340),
                      '5.8': grade_prop(5, 24, 380, 520, 420),
                      '8.8': grade_prop(16, 36, 600, 830, 660),
                      '9.8': grade_prop(1.6, 16, 650, 900, 720),
                      '10.9': grade_prop(5, 36, 830, 1040, 940),
                      '12.9': grade_prop(1.6, 36, 970, 1220, 1100)}
        if not isinstance(grade, str):
            raise KeyError(f"The grade should be a string")
        if diameter < grade_list[grade].low or diameter > grade_list[grade].high:
            raise Exception(f"Diameter({diameter}) not in range([{grade_list[grade].low},"
                            f"{grade_list[grade].high}]) for this grade({grade})")
        else:
            return [grade_list[grade].Sy, grade_list[grade].Sut, grade_list[grade].Sp]
